__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
from typing import Optional

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTFEncoder
from jina.excepts import PretrainedModelFileDoesNotExist


class CustomKerasImageEncoder(BaseTFEncoder):
    """
    :class:`CustomImageKerasEncoder` encodes data from a ndarray, potentially B x (Channel x Height x Width) into a
        ndarray of `B x D`.
    Internally, :class:`CustomImageKerasEncoder` wraps any custom tf.keras model not part of models from `tensorflow.keras.applications`.
    https://www.tensorflow.org/api_docs/python/tf/keras/applications
    """

    def __init__(self,
                 model_path: Optional[str] = 'models/mobilenet_v2.h5',
                 layer_name: Optional[str] = 'global_average_pooling2d',
                 channel_axis: int = 1,
                 *args, **kwargs):

        """
        :param model_path: the path where the model is stored.
        :layer: Name of the layer from where to extract the feature map.
        """
        super().__init__(*args, **kwargs)
        self.model_path = model_path
        self.layer_name = layer_name
        self.channel_axis = channel_axis

    def post_init(self):
        self.to_device()
        if self.model_path and os.path.exists(self.model_path):
            import tensorflow as tf
            model = tf.keras.models.load_model(self.model_path)
            model.trainable = False
            self.model = tf.keras.Model(inputs=model.input,
                                        outputs=model.get_layer(self.layer_name).output)
        else:
            raise PretrainedModelFileDoesNotExist(f'model {self.model_path} does not exist')

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param data: a `B x (Channel x Height x Width)` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``, `D` is the output dimension
        """
        if self.channel_axis != -1:
            data = np.moveaxis(data, self.channel_axis, -1)
        return self.model(data)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Dict, Any

import numpy as np

from . import BaseRecursiveDriver

if False:
    from ..types.document import Document
    from ..types.sets import DocumentSet


class ReduceAllDriver(BaseRecursiveDriver):
    """:class:`ReduceAllDriver` merges chunks/matches from all requests, recursively.

    .. note::

        It uses the last request as a reference.
    """

    def __init__(self, traversal_paths: Tuple[str] = ('c',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        # all pointers of the docs, provide the weak ref to all docs in partial reqs
        self.doc_pointers = {}  # type: Dict[str, Any]
        self._traverse_apply(self.docs, *args, **kwargs)
        self.doc_pointers.clear()

    def _apply_root(self, docs: 'DocumentSet', field: str, *args, **kwargs):
        docs = []
        for doc in self.docs:
            docs.append(doc)
        request = self.msg.request
        request.body.ClearField(field)
        request.docs.extend(docs)

    def _apply_all(
            self,
            docs: 'DocumentSet',
            context_doc: 'Document',
            field: str,
            *args,
            **kwargs) -> None:

        if context_doc.id not in self.doc_pointers:
            self.doc_pointers[context_doc.id] = context_doc
        else:
            getattr(self.doc_pointers[context_doc.id], field).extend(docs)


class CollectEvaluationDriver(ReduceAllDriver):
    """Merge all evaluations into one, grouped by ``doc.id`` """

    def _apply_all(
            self,
            docs: 'DocumentSet',
            context_doc: 'Document',
            field: str,
            *args,
            **kwargs) -> None:
        if context_doc.id not in self.doc_pointers:
            self.doc_pointers[context_doc.id] = context_doc.evaluations
        else:
            self.doc_pointers[context_doc.id].extend(context_doc.evaluations)


class ConcatEmbedDriver(ReduceAllDriver):
    """Concat all embeddings into one, grouped by ```doc.id``` """

    def __call__(self, *args, **kwargs):
        # all pointers of the docs, provide the weak ref to all docs in partial reqs
        self.doc_pointers = {}  # type: Dict[str, Any]
        self._traverse_apply(self.docs, *args, **kwargs)
        self._traverse_apply(self.req.docs, concatenate=True, *args, **kwargs)
        self.doc_pointers.clear()

    def _apply_all(
            self,
            docs: 'DocumentSet',
            context_doc: 'Document',
            field: str,
            concatenate: bool = False,
            *args,
            **kwargs):
        doc = context_doc
        if concatenate:
            doc.embedding = np.concatenate(self.doc_pointers[doc.id], axis=0)
        else:
            if doc.id not in self.doc_pointers:
                self.doc_pointers[doc.id] = [doc.embedding]
            else:
                self.doc_pointers[doc.id].append(doc.embedding)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BasePaddleEncoder


class VideoPaddleEncoder(BasePaddleEncoder):
    """
    :class:`VideoPaddleEncoder` encodes data from a ndarray, potentially B x T x (Channel x Height x Width) into a
        ndarray of `B x D`.
    Internally, :class:`VideoPaddleEncoder` wraps the models from `paddlehub`.
    https://github.com/PaddlePaddle/PaddleHub
    """

    def __init__(self,
                 model_name: str = 'tsn_kinetics400',
                 output_feature: str = '@HUB_tsn_kinetics400@reduce_mean_0.tmp_0',
                 pool_strategy: str = None,
                 channel_axis: int = 2,
                 *args, **kwargs):
        """

        :param model_name: the name of the model. Supported models include ``tsn_kinetics400``, ``stnet_kinetics400``,
            ``tsm_kinetics400``
        :param output_feature: the name of the layer for feature extraction. Please use the following values for the
            supported models:
            ``tsn_kinetics400``: `@HUB_tsn_kinetics400@reduce_mean_0.tmp_0`
            ``stnet_kinetics400``: ``@HUB_stnet_kinetics400@reshape2_6.tmp_0``
            ``tsm_kinetics400``: ``@HUB_tsm_kinetics400@reduce_mean_0.tmp_0``
        :param pool_strategy: the pooling strategy
            - `None` means that the output of the model will be the output feature.
            - `mean` means that global average pooling will be applied to the output feature, and thus the output of the
                model will be a 2D tensor.
            - `max` means that global max pooling will be applied.
        """
        super().__init__(*args, **kwargs)
        self.model_name = model_name
        self.outputs_name = output_feature
        self.pool_strategy = pool_strategy
        self.channel_axis = channel_axis
        self._default_channel_axis = 2
        if pool_strategy not in ('mean', 'max', None):
            raise NotImplementedError(f'unknown pool_strategy: {pool_strategy}')

    def post_init(self):
        import paddlehub as hub
        module = hub.Module(name=self.model_name)
        inputs, outputs, self.model = module.context(trainable=False)
        self.inputs_name = inputs[0].name
        self.exe = self.to_device()

    def close(self):
        self.exe.close()
        super().close()

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param data: a `B x T x (Channel x Height x Width)` numpy ``ndarray``, `B` is the size of the batch, `T` is the
            number of frames
        :return: a `B x D` numpy ``ndarray``, `D` is the output dimension
        """
        if self.channel_axis != self._default_channel_axis:
            data = np.moveaxis(data, self.channel_axis, self._default_channel_axis)
        feature_map, *_ = self.exe.run(
            program=self.model,
            fetch_list=[self.outputs_name],
            feed={self.inputs_name: data.astype('float32')},
            return_numpy=True
        )
        if feature_map.ndim == 2 or self.pool_strategy is None:
            return feature_map
        return self.get_pooling(feature_map)

    def get_pooling(self, data: 'np.ndarray') -> 'np.ndarray':
        _reduce_axis = tuple((i for i in range(len(data.shape)) if i > 1))
        return getattr(np, self.pool_strategy)(data, axis=_reduce_axis)



from collections.abc import MutableSequence
from typing import Iterable, Union, Dict


try:
    # when protobuf using Cpp backend
    from google.protobuf.pyext._message import RepeatedCompositeContainer as RepeatedContainer
except:
    # when protobuf using Python backend
    from google.protobuf.internal.containers import RepeatedCompositeFieldContainer as RepeatedContainer


from ..querylang import QueryLang
from ...helper import typename
from ...proto.jina_pb2 import QueryLangProto

AcceptQueryLangType = Union[QueryLang, QueryLangProto, Dict]

__all__ = ['QueryLangSet', 'AcceptQueryLangType']


class QueryLangSet(MutableSequence):
    """:class:`QueryLangSet` is a mutable sequence of :class:`QueryLang`,
    it gives an efficient view of a list of Document. One can iterate over it like
    a generator but ALSO modify it, count it, get item.
    """

    def __init__(self, querylang_protos: 'RepeatedContainer'):
        super().__init__()
        self._querylangs_proto = querylang_protos
        self._querylangs_map = {}

    def insert(self, index: int, ql: 'QueryLang') -> None:
        self._querylangs_proto.insert(index, ql.proto)

    def __setitem__(self, key, value: 'QueryLang'):
        if isinstance(key, int):
            self._querylangs_proto[key].CopyFrom(value.proto)
        elif isinstance(key, str):
            self._querylangs_map[key].CopyFrom(value.proto)
        else:
            raise IndexError(f'do not support this index {key}')

    def __delitem__(self, index):
        del self._querylangs_proto[index]

    def __len__(self):
        return len(self._querylangs_proto)

    def __iter__(self):
        for d in self._querylangs_proto:
            yield QueryLang(d)

    def __getitem__(self, item):
        if isinstance(item, int):
            return QueryLang(self._querylangs_proto[item])
        elif isinstance(item, str):
            return QueryLang(self._querylangs_map[item])
        else:
            raise IndexError(f'do not support this index {item}')

    def append(self, value: 'AcceptQueryLangType'):
        q_pb = self._querylangs_proto.add()
        if isinstance(value, Dict):
            q_pb.CopyFrom(QueryLang(value).proto)
        elif isinstance(value, QueryLangProto):
            q_pb.CopyFrom(value)
        elif isinstance(value, QueryLang):
            q_pb.CopyFrom(value.proto)
        else:
            raise TypeError(f'unknown type {typename(value)}')

    def extend(self, iterable: Iterable[AcceptQueryLangType]) -> None:
        for q in iterable:
            self.append(q)

    def clear(self):
        del self._querylangs_proto[:]

    def reverse(self):
        size = len(self._querylangs_proto)
        hi_idx = size - 1
        for i in range(int(size / 2)):
            tmp = QueryLangProto()
            tmp.CopyFrom(self._querylangs_proto[hi_idx])
            self._querylangs_proto[hi_idx].CopyFrom(self._querylangs_proto[i])
            self._querylangs_proto[i].CopyFrom(tmp)
            hi_idx -= 1

    def build(self):
        """Build a name to QueryLang mapping so one can later index a QueryLang using
        name as string key
        """
        # TODO This is a temp fix, QueryLangProto do not have an id field.
        self._querylangs_map = {q.name: q for q in self._querylangs_proto}



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict, Any

from .queryset.lookup import Q
from .. import QuerySetReader, BaseRecursiveDriver

if False:
    from ...types.sets import DocumentSet


class FilterQL(QuerySetReader, BaseRecursiveDriver):
    """Filters incoming `docs` by evaluating a series of `lookup rules`.

        This is often useful when the proceeding Pods require only a signal, not the full message.

        Example ::
        - !FilterQL
            with:
                lookups: {modality: mode2}
        - !EncodeDriver
            with:
                method: encode

        ensures that the EncodeDriver will only get documents which modality field value is `mode2` by filtering
        those documents at the specific levels that do not comply with this condition
    """

    def __init__(self, lookups: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        """
        :param lookups: (dict) a dictionary where keys are interpreted by ``:class:`LookupLeaf`` to form a
        an evaluation function. For instance, a dictionary ``{ modality__in: [mode1, mode2] }``, would create
        an evaluation function that will check if the field `modality` is found in `[mode1, mode2]`
        """
        self._lookups = lookups

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        if self.lookups:
            _lookups = Q(**self.lookups)
            miss_idx = []
            for idx, doc in enumerate(docs):
                if not _lookups.evaluate(doc):
                    miss_idx.append(idx)

            # delete non-exit matches in reverse
            for j in reversed(miss_idx):
                del docs[j]



from typing import Sequence, Union, Optional

from jina.executors.evaluators.rank import BaseRankingEvaluator


class ReciprocalRankEvaluator(BaseRankingEvaluator):
    """
    :class:`ReciprocalRankEvaluator` Gives score as per reciprocal rank metric.
    """

    def __init__(self, eval_at: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_at = eval_at

    def evaluate(self, actual: Sequence[Union[str, int]], desired: Sequence[Union[str, int]], *args, **kwargs) -> float:
        """
        :param actual: should be a sequence of document IDs
        :param desired: should be a sequence of document IDs
        :return gives reciprocal rank score
        """
        if len(actual) == 0 or len(desired) == 0:
            return 0.0
        try:
            actual_at_k = actual[:self.eval_at]
            return 1.0 / (actual_at_k.index(desired[0]) + 1)
        except:
            return 0.0



import argparse
import asyncio
from typing import Optional

from .client import PeaDaemonClient
from ..asyncio.base import AsyncZMQRuntime
from ...zmq import Zmqlet
from ....excepts import DaemonConnectivityError
from ....helper import cached_property, colored, is_yaml_filepath


class JinadRuntime(AsyncZMQRuntime):

    def __init__(self, args: 'argparse.Namespace'):
        super().__init__(args)
        self.ctrl_addr = Zmqlet.get_ctrl_address(None, None, True)[0]
        self.timeout_ctrl = args.timeout_ctrl
        self.host = args.host
        self.port_expose = args.port_expose
        self.api = PeaDaemonClient(host=self.host, port=self.port_expose, logger=self.logger,
                                   timeout=self.args.timeout_ready)

    def setup(self):
        """
        Uploads Pod/Pea context to remote & Creates remote Pod/Pea using :class:`JinadAPI`
        """
        if self._remote_id:
            self.logger.success(f'created a remote {self.api.kind}: {colored(self._remote_id, "cyan")}')
        else:
            self.logger.error(
                f'fail to connect to the daemon at {self.host}:{self.port_expose}, please check:\n'
                f'- is there a typo in {self.host}?\n'
                f'- on {self.host}, are you running `docker run --network host jinaai/jina:latest-daemon`?\n'
                f'- on {self.host}, have you set the security policy to public for all required ports?\n'
                f'- on local, are you behind VPN or proxy?')
            raise DaemonConnectivityError

    async def async_run_forever(self):
        """
        Streams log messages using websocket from remote server
        """
        self._logging_task = asyncio.create_task(
            self._sleep_forever() if self.args.silent_remote_logs else
            self.api.logstream(self._workspace_id, self._remote_id)
        )

    async def async_cancel(self):
        """
        Cancels the logging task
        """
        self._logging_task.cancel()

    def teardown(self):
        """
        Closes the remote Pod/Pea using :class:`JinadAPI`
        """
        self.api.delete(remote_id=self._remote_id)

    @cached_property
    def _remote_id(self) -> Optional[str]:
        if self.api.is_alive:
            upload_files = []
            if is_yaml_filepath(self.args.uses):
                upload_files.append(self.args.uses)

            if is_yaml_filepath(self.args.uses_internal):
                upload_files.append(self.args.uses_internal)

            if self.args.upload_files:
                upload_files.extend(self.args.upload_files)
            else:
                self.logger.warning(f'will upload {upload_files} to remote, to include more local file '
                                    f'dependencies, please use `--upload-files`')

            if upload_files:
                workspace_id = self.api.upload(dependencies=list(set(upload_files)),
                                               workspace_id=self.args.workspace_id)
                if workspace_id:
                    self.logger.success(f'uploaded to workspace: {colored(workspace_id, "cyan")}')
                else:
                    raise RuntimeError('can not upload required files to remote')

            _id = self.api.create(self.args)

            # if there is a new workspace_id, then use it
            self._workspace_id = self.api.get_status(_id)['workspace_id']
            return _id

    async def _sleep_forever(self):
        """Sleep forever, no prince will come.
        """
        await asyncio.sleep(1e10)



from typing import Tuple, Sequence

from ... import Document, Request
from ...enums import DataInputType, RequestType
from ...excepts import BadDocType, BadRequestType


def _new_doc_from_data(data, data_type: DataInputType, **kwargs) -> Tuple['Document', 'DataInputType']:
    def _build_doc_from_content():
        with Document(**kwargs) as d:
            d.content = data
        return d, DataInputType.CONTENT

    if data_type == DataInputType.AUTO or data_type == DataInputType.DOCUMENT:
        if isinstance(data, Document):
            # if incoming is already primitive type Document, then all good, best practice!
            return data, DataInputType.DOCUMENT
        try:
            d = Document(data, **kwargs)
            return d, DataInputType.DOCUMENT
        except BadDocType:
            # AUTO has a fallback, now reconsider it as content
            if data_type == DataInputType.AUTO:
                return _build_doc_from_content()
            else:
                raise
    elif data_type == DataInputType.CONTENT:
        return _build_doc_from_content()


def _new_request_from_batch(_kwargs, batch, data_type, mode, queryset):
    req = Request()
    req.request_type = str(mode)

    try:
        # add type-specific fields
        if mode == RequestType.INDEX or mode == RequestType.SEARCH or mode == RequestType.TRAIN or mode == RequestType.UPDATE:
            _add_docs_groundtruths(req, batch, data_type, _kwargs)
        elif mode == RequestType.DELETE:
            _add_ids(req, batch)
        else:
            raise NotImplementedError(f'generating request from {mode} is not yet supported')
    except Exception as ex:
        raise BadRequestType(f'error when building {req.request_type} from {batch}') from ex

    # add common fields
    if isinstance(queryset, Sequence):
        req.queryset.extend(queryset)
    elif queryset is not None:
        req.queryset.append(queryset)
    return req


def _add_docs_groundtruths(req, batch, data_type, _kwargs):
    for content in batch:
        if isinstance(content, tuple) and len(content) == 2:
            # content comes in pair,  will take the first as the input and the second as the groundtruth

            # note how data_type is cached
            d, data_type = _new_doc_from_data(content[0], data_type, **_kwargs)
            gt, _ = _new_doc_from_data(content[1], data_type, **_kwargs)
            req.docs.append(d)
            req.groundtruths.append(gt)
        else:
            d, data_type = _new_doc_from_data(content, data_type, **_kwargs)
            req.docs.append(d)


def _add_ids(req, batch):
    string_ids = (str(doc_id) for doc_id in batch)
    req.ids.extend(string_ids)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Any, Iterator

import numpy

from . import BaseExecutableDriver
from .querylang.queryset.dunderkey import dunder_get
from .search import KVSearchDriver
from ..types.document import Document
from ..types.document.helper import DocGroundtruthPair


class BaseEvaluateDriver(BaseExecutableDriver):
    def __init__(self, executor: str = None,
                 method: str = 'evaluate',
                 running_avg: bool = False,
                 *args,
                 **kwargs):
        """
        :param executor: the name of the sub-executor, only necessary when :class:`jina.executors.compound.CompoundExecutor` is used
        :param method: the function name of the executor that the driver feeds to
        :param running_avg: always return running average instead of value of the current run
        :param args:
        :param kwargs:

        .. warning::

            When ``running_avg=True``, then the running mean is returned. So far at Jina 0.8.10,
             there is no way to reset the running statistics. If you have a query Flow running multiple queries,
             you may want to make sure the running statistics is meaningful across multiple runs.
        """
        super().__init__(executor, method, *args, **kwargs)
        self._running_avg = running_avg

    def __call__(self, *args, **kwargs):
        docs_groundtruths = [DocGroundtruthPair(doc, groundtruth) for doc, groundtruth in
                             zip(self.docs, self.req.groundtruths)]
        self._traverse_apply(docs_groundtruths, *args, **kwargs)

    def _apply_all(
            self,
            docs: Iterator['DocGroundtruthPair'],
            *args,
            **kwargs
    ) -> None:
        for doc_groundtruth in docs:
            doc = doc_groundtruth.doc
            groundtruth = doc_groundtruth.groundtruth
            evaluation = doc.evaluations.add()
            evaluation.value = self.exec_fn(self.extract(doc), self.extract(groundtruth))
            if self._running_avg:
                evaluation.value = self.exec.mean

            if getattr(self.exec, 'eval_at', None):
                evaluation.op_name = f'{self.exec.__class__.__name__}@{self.exec.eval_at}'
            else:
                evaluation.op_name = self.exec.__class__.__name__
            evaluation.ref_id = groundtruth.id

    def extract(self, doc: 'Document') -> Any:
        """Extracting the to-be-evaluated field from the document.
        Drivers inherit from :class:`BaseEvaluateDriver` must implement this method.

        This function will be invoked two times in :meth:`_apply_all`:
        once with actual doc, once with groundtruth doc.
        """
        raise NotImplementedError


class FieldEvaluateDriver(BaseEvaluateDriver):
    """
    Evaluate on the values from certain field, the extraction is implemented with :meth:`dunder_get`
    """

    def __init__(self, field: str,
                 *args,
                 **kwargs):
        """

        :param field: the field name to be extracted from the Protobuf
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.field = field

    def extract(self, doc: 'Document') -> Any:
        return dunder_get(doc, self.field)


class RankEvaluateDriver(FieldEvaluateDriver):
    """Drivers used to pass `matches` from documents and groundtruths to an executor and add the evaluation value

        - Example fields:
        ['tags__id', 'id', 'score__value]
    """

    def __init__(self,
                 field: str = 'tags__id',
                 *args,
                 **kwargs):
        """
        :param args:
        :param kwargs:
        """
        super().__init__(field, *args, **kwargs)

    def extract(self, doc: 'Document'):
        r = [dunder_get(x, self.field) for x in doc.matches]
        # flatten nested list but useless depth, e.g. [[1,2,3,4]]
        return list(numpy.array(r).flat)


class NDArrayEvaluateDriver(FieldEvaluateDriver):
    """Drivers used to pass `embedding` from documents and groundtruths to an executor and add the evaluation value

    - Valid fields:
        ['blob', 'embedding']

    """

    def __init__(self, field: str = 'embedding', *args, **kwargs):
        super().__init__(field, *args, **kwargs)


class TextEvaluateDriver(FieldEvaluateDriver):
    """Drivers used to pass a content field from documents and groundtruths to an executor and add the evaluation value

    - Valid fields:
                ['id',
                 'level_name',
                 'parent_id',
                 'text',
                 'mime_type',
                 'uri',
                 'modality']
    """

    def __init__(self, field: str = 'text', *args, **kwargs):
        super().__init__(field, *args, **kwargs)


class LoadGroundTruthDriver(KVSearchDriver):
    """Driver used to search for the `document key` in a KVIndex to find the corresponding groundtruth.
     (This driver does not use the `recursive structure` of jina Documents, and will not consider the `traversal_path` argument.
     It only retrieves `groundtruth` taking documents at root as key)
     This driver's job is to fill the `request` groundtruth with the corresponding groundtruth for each document if found in the corresponding KVIndexer.

    .. warning::
        The documents that are not found to have an indexed groundtruth are removed from the `request` so that the `Evaluator` only
        works with documents which have groundtruth.
    """

    def __call__(self, *args, **kwargs):
        miss_idx = []  #: missed hit results, some documents may not have groundtruth and thus will be removed
        for idx, doc in enumerate(self.docs):
            serialized_groundtruth = self.exec_fn(doc.id)
            if serialized_groundtruth:
                self.req.groundtruths.append(Document(serialized_groundtruth))
            else:
                miss_idx.append(idx)
        # delete non-existed matches in reverse
        for j in reversed(miss_idx):
            del self.docs[j]



from .. import BaseExecutor

if False:
    import numpy as np


class BaseClassifier(BaseExecutor):
    """
    The base class of Classifier Executor. Classifier Executor allows one to
    perform classification and regression on given input and output the predicted
    hard/soft label.

    This class should not be used directly. Subclasses should be used.
    """

    def predict(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """ Perform hard/soft classification on ``data``, the predicted value for each sample in X is returned.

        The output value can be zero/one, for one-hot label; or float for soft-label or regression label.
        Use the corresponding driver to interpret these labels

        The size and type of output can be one of the follows, ``B`` is ``data.shape[0]``:
            - (B,) or (B, 1); zero/one or float
            - (B, L): zero/one one-hot or soft label for L-class multi-class classification

        :param data: the input data to be classified, can be a ndim array.
            where axis=0 represents the batch size, i.e. data[0] is the first sample, data[1] is the second sample, ...
        :param args:
        :param kwargs:
        :return: the predicted value of each sample.
        """
        raise NotImplementedError
